@@ -229,11 +229,9 @@ int main(argc, argv)
 	    }
 
        } else if (strcmp(*argv, "-x") == 0 && ARG_VAL) {
-	   db5util_db_args_size++;
 	   if (!add_db_arg(koptarg)) {
-	       fprintf(stderr,"%s: Cannot initialize. Not enough memory\n",
-		       argv[0]);
-	       exit(1);
+		com_err(progname, ENOMEM, "while parsing command arguments\n");
+		exit(1);
 	   }
 
        } else if (strcmp(*argv, "-r") == 0 && ARG_VAL) {
@@ -380,7 +378,7 @@ static int open_db_and_mkey()
     valid_master_key = 0;
 
     if ((retval = krb5_db_open(util_context, db5util_db_args, 
-			       KRB5_KDB_OPEN_RW | KRB5_KDB_SRV_TYPE_OTHER))) {
+			       KRB5_KDB_OPEN_RW | KRB5_KDB_SRV_TYPE_ADMIN))) {
 	com_err(progname, retval, "while initializing database");
 	exit_status++;
 	return(1);
