@@ -301,15 +301,6 @@ int main(int argc, char *argv[])
 
      krb5_klog_init(context, "admin_server", whoami, 1);
 
-     krb5_klog_syslog(LOG_INFO, "Seeding random number generator");
-          ret = krb5_c_random_os_entropy(context, 1, NULL);
-	  if(ret) {
-	    krb5_klog_syslog(LOG_ERR,
-			     "Error getting random seed: %s, aborting",
-			     krb5_get_error_message (context, ret));
-	    exit(1);
-	  }
-	  
      if((ret = kadm5_init("kadmind", NULL,
 			  NULL, &params,
 			  KADM5_STRUCT_VERSION,
@@ -318,7 +309,7 @@ int main(int argc, char *argv[])
 		     &global_server_handle)) != KADM5_OK) {
 	  const char *e_txt = krb5_get_error_message (context, ret);
 	  krb5_klog_syslog(LOG_ERR, "%s while initializing, aborting",
-		 e_txt);
+			   e_txt);
 	  fprintf(stderr, "%s: %s while initializing, aborting\n",
 		  whoami, e_txt);
 	  krb5_klog_close(context);
@@ -639,6 +630,17 @@ kterr:
 	  exit(1);
      }
      
+     krb5_klog_syslog(LOG_INFO, "Seeding random number generator");
+     ret = krb5_c_random_os_entropy(context, 1, NULL);
+     if (ret) {
+	  krb5_klog_syslog(LOG_ERR, "Error getting random seed: %s, aborting",
+			   krb5_get_error_message(context, ret));
+	  svcauth_gssapi_unset_names();
+	  kadm5_destroy(global_server_handle);
+	  krb5_klog_close(context);
+	  exit(1);
+     }
+	  
      setup_signal_handlers();
      krb5_klog_syslog(LOG_INFO, "starting");
      kadm_svc_run(&params);
@@ -847,6 +849,7 @@ void reset_db(void)
 {
 #ifdef notdef
      kadm5_ret_t ret;
+     char *errmsg;
      
      if (ret = kadm5_flush(global_server_handle)) {
 	  krb5_klog_syslog(LOG_ERR, "FATAL ERROR!  %s while flushing databases.  "
@@ -989,8 +992,6 @@ void log_badverf(gss_name_t client_name, gss_name_t server_name,
      rpcproc_t proc;
      int i;
      const char *procname;
-     size_t clen, slen;
-     char *cdots, *sdots;
 
      client.length = 0;
      client.value = NULL;
@@ -999,20 +1000,10 @@ void log_badverf(gss_name_t client_name, gss_name_t server_name,
 
      (void) gss_display_name(&minor, client_name, &client, &gss_type);
      (void) gss_display_name(&minor, server_name, &server, &gss_type);
-     if (client.value == NULL) {
+     if (client.value == NULL)
 	 client.value = "(null)";
-	 clen = sizeof("(null)") -1;
-     } else {
-	 clen = client.length;
-     }
-     trunc_name(&clen, &cdots);
-     if (server.value == NULL) {
+     if (server.value == NULL)
 	 server.value = "(null)";
-	 slen = sizeof("(null)") - 1;
-     } else {
-	 slen = server.length;
-     }
-     trunc_name(&slen, &sdots);
      a = inet_ntoa(rqst->rq_xprt->xp_raddr.sin_addr);
 
      proc = msg->rm_call.cb_proc;
@@ -1025,14 +1016,14 @@ void log_badverf(gss_name_t client_name, gss_name_t server_name,
      }
      if (procname != NULL)
 	  krb5_klog_syslog(LOG_NOTICE, "WARNING! Forged/garbled request: %s, "
-			   "claimed client = %.*s%s, server = %.*s%s, addr = %s",
-			   procname, clen, client.value, cdots,
-			   slen, server.value, sdots, a);
+			   "claimed client = %s, server = %s, addr = %s",
+			   procname, client.value,
+			   server.value, a);
      else
 	  krb5_klog_syslog(LOG_NOTICE, "WARNING! Forged/garbled request: %d, "
-			   "claimed client = %.*s%s, server = %.*s%s, addr = %s",
-			   proc, clen, client.value, cdots,
-			   slen, server.value, sdots, a);
+			   "claimed client = %s, server = %s, addr = %s",
+			   proc, client.value,
+			   server.value, a);
 
      (void) gss_release_buffer(&minor, &client);
      (void) gss_release_buffer(&minor, &server);
@@ -1181,7 +1172,7 @@ void do_schpw(int s1, kadm5_config_params *params)
        interoperate if the client is single-homed. */
 
     if ((s2 = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
-	char *errmsg = krb5_get_error_message (context, errno);
+	const char *errmsg = krb5_get_error_message (context, errno);
 	krb5_klog_syslog(LOG_ERR, "cannot create connecting socket: %s",
 			 errmsg);
 	fprintf(stderr, "Cannot create connecting socket: %s",
