@@ -51,8 +51,8 @@
  * organization.
  */
 #define KRB5_MAJOR_RELEASE 1
-#define KRB5_MINOR_RELEASE 5
-#define KRB5_PATCHLEVEL 4
+#define KRB5_MINOR_RELEASE 6
+#define KRB5_PATCHLEVEL 0
 /* #undef KRB5_RELTAIL */
-#define KRB5_RELDATE "20070706"
-#define KRB5_RELTAG "tags/krb5-1-5-4-final"
+#define KRB5_RELDATE "20070109"
+#define KRB5_RELTAG "tags/krb5-1-6-final"
