@@ -24,6 +24,7 @@
 
 /* $Id$ */
 
+#include<shlwapi.h>
 #include<kconfiginternal.h>
 #include<assert.h>
 
@@ -63,6 +64,9 @@ void exit_kconf(void) {
 
         khcint_free_space(conf_root);
 
+        LeaveCriticalSection(&cs_conf_global);
+        DeleteCriticalSection(&cs_conf_global);
+
         EnterCriticalSection(&cs_conf_handle);
         while(conf_free_handles) {
             LPOP(&conf_free_handles, &h);
@@ -79,12 +83,10 @@ void exit_kconf(void) {
         }
         LeaveCriticalSection(&cs_conf_handle);
         DeleteCriticalSection(&cs_conf_handle);
-
-        LeaveCriticalSection(&cs_conf_global);
-        DeleteCriticalSection(&cs_conf_global);
     }
 }
 
+/* obtains cs_conf_handle/cs_conf_global */
 kconf_handle * 
 khcint_handle_from_space(kconf_conf_space * s, khm_int32 flags)
 {
@@ -109,7 +111,7 @@ khcint_handle_from_space(kconf_conf_space * s, khm_int32 flags)
     return h;
 }
 
-/* must be called with cs_conf_global held */
+/* obtains cs_conf_handle/cs_conf_global */
 void 
 khcint_handle_free(kconf_handle * h)
 {
@@ -129,6 +131,10 @@ khcint_handle_free(kconf_handle * h)
 
         if(a == NULL) {
             DebugBreak();
+
+            /* hmm.  the handle was not in the in-use list */
+            LeaveCriticalSection(&cs_conf_handle);
+            return;
         }
     }
 #endif
@@ -139,12 +145,14 @@ khcint_handle_free(kconf_handle * h)
             h->space = NULL;
         }
         lower = h->lower;
+        h->magic = 0;
         LPUSH(&conf_free_handles, h);
         h = lower;
     }
     LeaveCriticalSection(&cs_conf_handle);
 }
 
+/* obains cs_conf_handle/cs_conf_global */
 kconf_handle * 
 khcint_handle_dup(kconf_handle * o)
 {
@@ -164,6 +172,7 @@ khcint_handle_dup(kconf_handle * o)
     return r;
 }
 
+/* obtains cs_conf_global */
 void 
 khcint_space_hold(kconf_conf_space * s) {
     EnterCriticalSection(&cs_conf_global);
@@ -171,6 +180,7 @@ khcint_space_hold(kconf_conf_space * s) {
     LeaveCriticalSection(&cs_conf_global);
 }
 
+/* obtains cs_conf_global */
 void 
 khcint_space_release(kconf_conf_space * s) {
     khm_int32 l;
@@ -314,6 +324,11 @@ khcint_RegOpenKeyEx(HKEY hkey, LPCWSTR sSubKey, DWORD ulOptions,
     return rv;
 }
 
+/*! \internal
+
+ \note This function is not a good replacement for RegDeleteKey since
+     it deletes all the subkeys in addition to the key being deleted.
+ */
 LONG
 khcint_RegDeleteKey(HKEY hKey,
                     LPCWSTR lpSubKey) {
@@ -345,7 +360,15 @@ khcint_RegDeleteKey(HKEY hKey,
             /* bingo! ?? */
             if ((sk_name[cch] == L'\0' ||
                  sk_name[cch] == L'~')) {
-                rv = RegDeleteKey(hKey, sk_name);
+
+                /* instead of calling RegDeleteKey we call SHDeleteKey
+                   because we want to blow off all the subkeys as
+                   well.  This is different from the behavior of
+                   RegDeleteKey making khcint_RegDeleteKey not a very
+                   good case sensitive replacement for
+                   RegDeleteKey. */
+
+                rv = SHDeleteKey(hKey, sk_name);
                 goto _cleanup;
             }
         }
@@ -521,7 +544,7 @@ khcint_RegCreateKeyEx(HKEY hKey,
     return rv;
 }
 
-
+/* obtains cs_conf_global */
 HKEY 
 khcint_space_open_key(kconf_conf_space * s, khm_int32 flags) {
     HKEY hk = NULL;
@@ -591,6 +614,7 @@ khcint_space_open_key(kconf_conf_space * s, khm_int32 flags) {
     }
 }
 
+/* obtains cs_conf_handle/cs_conf_global */
 KHMEXP khm_int32 KHMAPI 
 khc_shadow_space(khm_handle upper, khm_handle lower)
 {
@@ -599,18 +623,20 @@ khc_shadow_space(khm_handle upper, khm_handle lower)
     if(!khc_is_config_running())
         return KHM_ERROR_NOT_READY;
 
-    if(!khc_is_handle(upper))
+    if(!khc_is_handle(upper)) {
+#ifdef DEBUG
+        DebugBreak();
+#endif
         return KHM_ERROR_INVALID_PARAM;
+    }
 
     h = (kconf_handle *) upper;
 
     EnterCriticalSection(&cs_conf_handle);
     if(h->lower) {
-        LeaveCriticalSection(&cs_conf_handle);
         EnterCriticalSection(&cs_conf_global);
         khcint_handle_free(h->lower);
         LeaveCriticalSection(&cs_conf_global);
-        EnterCriticalSection(&cs_conf_handle);
         h->lower = NULL;
     }
 
@@ -619,9 +645,7 @@ khc_shadow_space(khm_handle upper, khm_handle lower)
         kconf_handle * lc;
 
         l = (kconf_handle *) lower;
-        LeaveCriticalSection(&cs_conf_handle);
         lc = khcint_handle_dup(l);
-        EnterCriticalSection(&cs_conf_handle);
         h->lower = lc;
     }
     LeaveCriticalSection(&cs_conf_handle);
@@ -629,6 +653,7 @@ khc_shadow_space(khm_handle upper, khm_handle lower)
     return KHM_ERROR_SUCCESS;
 }
 
+/* no locks */
 kconf_conf_space * 
 khcint_create_empty_space(void) {
     kconf_conf_space * r;
@@ -640,6 +665,7 @@ khcint_create_empty_space(void) {
     return r;
 }
 
+/* called with cs_conf_global */
 void 
 khcint_free_space(kconf_conf_space * r) {
     kconf_conf_space * c;
@@ -668,6 +694,7 @@ khcint_free_space(kconf_conf_space * r) {
     PFREE(r);
 }
 
+/* obtains cs_conf_global */
 khm_int32 
 khcint_open_space(kconf_conf_space * parent, 
                   const wchar_t * sname, size_t n_sname, 
@@ -677,6 +704,7 @@ khcint_open_space(kconf_conf_space * parent,
     HKEY pkey = NULL;
     HKEY ckey = NULL;
     wchar_t buf[KCONF_MAXCCH_NAME];
+    size_t cb_regpath = 0;
 
     if(!parent)
         p = conf_root;
@@ -686,10 +714,7 @@ khcint_open_space(kconf_conf_space * parent,
     if(n_sname >= KCONF_MAXCCH_NAME || n_sname <= 0)
         return KHM_ERROR_INVALID_PARAM;
 
-    /* SAFE: buf: buffer size == KCONF_MAXCCH_NAME * wchar_t >
-       n_sname * wchar_t */
-    wcsncpy(buf, sname, n_sname);
-    buf[n_sname] = L'\0';
+    StringCchCopyN(buf, ARRAYLENGTH(buf), sname, n_sname);
 
     /* see if there is already a config space by this name. if so,
        return it.  Note that if the configuration space is specified
@@ -760,19 +785,18 @@ khcint_open_space(kconf_conf_space * parent,
     /*SAFE: p->regpath: is valid since it was set using this same
       function. */
     /*SAFE: buf: see above */
-    c->regpath = PMALLOC((wcslen(p->regpath) + wcslen(buf) + 2) * sizeof(wchar_t));
+    cb_regpath = (wcslen(p->regpath) + wcslen(buf) + 2) * sizeof(wchar_t);
+    c->regpath = PMALLOC(cb_regpath);
 
     assert(c->regpath != NULL);
 
-#pragma warning( push )
-#pragma warning( disable: 4995 )
     /*SAFE: c->regpath: allocated above to be big enough */
     /*SAFE: p->regpath: see above */
-    wcscpy(c->regpath, p->regpath);
-    wcscat(c->regpath, L"\\");
+    StringCbCopy(c->regpath, cb_regpath, p->regpath);
+    StringCbCat(c->regpath, cb_regpath, L"\\");
+
     /*SAFE: buf: see above */
-    wcscat(c->regpath, buf);
-#pragma warning( pop )
+    StringCbCat(c->regpath, cb_regpath, buf);
 
     khcint_space_hold(c);
 
@@ -784,6 +808,7 @@ khcint_open_space(kconf_conf_space * parent,
     return KHM_ERROR_SUCCESS;
 }
 
+/* obtains cs_conf_handle/cs_conf_global */
 KHMEXP khm_int32 KHMAPI 
 khc_open_space(khm_handle parent, const wchar_t * cspace, khm_int32 flags, 
                khm_handle * result) {
@@ -798,8 +823,12 @@ khc_open_space(khm_handle parent, const wchar_t * cspace, khm_int32 flags,
         return KHM_ERROR_NOT_READY;
     }
 
-    if(!result || (parent && !khc_is_handle(parent)))
+    if(!result || (parent && !khc_is_handle(parent))) {
+#ifdef DEBUG
+        DebugBreak();
+#endif
         return KHM_ERROR_INVALID_PARAM;
+    }
 
     if(!parent)
         p = conf_root;
@@ -837,11 +866,6 @@ khc_open_space(khm_handle parent, const wchar_t * cspace, khm_int32 flags,
 
             end = wcschr(str, L'\\'); /* safe because cspace was
                                      validated above */
-#if 0
-            if(!end)
-                end = wcschr(str, L'/');  /* safe because cspace was
-                                         validated above */
-#endif
         }
 
         if(!end) {
@@ -857,11 +881,7 @@ khc_open_space(khm_handle parent, const wchar_t * cspace, khm_int32 flags,
 
         rv = khcint_open_space(p, str, end - str, flags, &c);
 
-        if(KHM_SUCCEEDED(rv) && (*end == L'\\'
-#if 0
-                                 || *end == L'/'
-#endif
-                                 )) {
+        if(KHM_SUCCEEDED(rv) && (*end == L'\\')) {
             khcint_space_release(p);
             p = c;
             c = NULL;
@@ -880,21 +900,27 @@ khc_open_space(khm_handle parent, const wchar_t * cspace, khm_int32 flags,
     return rv;
 }
 
+/* obtains cs_conf_handle/cs_conf_global */
 KHMEXP khm_int32 KHMAPI 
 khc_close_space(khm_handle csp) {
     if(!khc_is_config_running())
         return KHM_ERROR_NOT_READY;
 
-    if(!khc_is_handle(csp))
+    if(!khc_is_handle(csp)) {
+#ifdef DEBUG
+        DebugBreak();
+#endif
         return KHM_ERROR_INVALID_PARAM;
+    }
 
     khcint_handle_free((kconf_handle *) csp);
     return KHM_ERROR_SUCCESS;
 }
 
+/* obtains cs_conf_handle/cs_conf_global */
 KHMEXP khm_int32 KHMAPI 
 khc_read_string(khm_handle pconf, 
-                wchar_t * pvalue, 
+                const wchar_t * pvalue, 
                 wchar_t * buf, 
                 khm_size * bufsize) 
 {
@@ -907,7 +933,7 @@ khc_read_string(khm_handle pconf,
     do {
         HKEY hku = NULL;
         HKEY hkm = NULL;
-        wchar_t * value = NULL;
+        const wchar_t * value = NULL;
         int free_space = 0;
         khm_handle conf = NULL;
         DWORD size;
@@ -916,11 +942,7 @@ khc_read_string(khm_handle pconf,
 
         int i;
 
-        if(wcschr(pvalue, L'\\')
-#if 0
-           || wcschr(pvalue, L'/')
-#endif
-           ) {
+        if((value = wcsrchr(pvalue, L'\\')) != NULL) {
 
             if(KHM_FAILED(khc_open_space(
                 pconf, 
@@ -930,15 +952,14 @@ khc_read_string(khm_handle pconf,
                 goto _shadow;
 
             free_space = 1;
-#if 0
-            wchar_t * back, * forward;
 
-            back = wcsrchr(pvalue, L'\\');
-            forward = wcsrchr(pvalue, L'/');
-            value = (back > forward)?back:forward; /* works for nulls too */
-#else
-            value = wcsrchr(pvalue, L'\\');
+            if (value) {
+                value++;
+            } else {
+#ifdef DEBUG
+                assert(FALSE);
 #endif
+            }
         } else {
             value = pvalue;
             conf = pconf;
@@ -1055,8 +1076,9 @@ _exit:
     return rv;
 }
 
+/* obtains cs_conf_handle/cs_conf_global */
 KHMEXP khm_int32 KHMAPI 
-khc_read_int32(khm_handle pconf, wchar_t * pvalue, khm_int32 * buf) {
+khc_read_int32(khm_handle pconf, const wchar_t * pvalue, khm_int32 * buf) {
     kconf_conf_space * c;
     khm_int32 rv = KHM_ERROR_SUCCESS;
 
@@ -1073,17 +1095,13 @@ khc_read_int32(khm_handle pconf, wchar_t * pvalue, khm_int32 * buf) {
         HKEY hku = NULL;
         HKEY hkm = NULL;
 
-        wchar_t * value = NULL;
+        const wchar_t * value = NULL;
         int free_space = 0;
         khm_handle conf = NULL;
 
         int i;
 
-        if(wcschr(pvalue, L'\\')
-#if 0
-           || wcschr(pvalue, L'/')
-#endif
-           ) {
+        if((value = wcsrchr(pvalue, L'\\')) != NULL) {
             if(KHM_FAILED(khc_open_space(
                 pconf, 
                 pvalue, 
@@ -1091,15 +1109,14 @@ khc_read_int32(khm_handle pconf, wchar_t * pvalue, khm_int32 * buf) {
                 &conf)))
                 goto _shadow;
             free_space = 1;
-#if 0
-            wchar_t * back, * forward;
 
-            back = wcsrchr(pvalue, L'\\');
-            forward = wcsrchr(pvalue, L'/');
-            value = (back > forward)?back:forward;
-#else
-            value = wcsrchr(pvalue, L'\\');
+            if (value) {
+                value++;
+            } else {
+#ifdef DEBUG
+                assert(FALSE);
 #endif
+            }
         } else {
             value = pvalue;
             conf = pconf;
@@ -1177,8 +1194,9 @@ _exit:
     return rv;
 }
 
+/* obtains cs_conf_handle/cs_conf_global */
 KHMEXP khm_int32 KHMAPI 
-khc_read_int64(khm_handle pconf, wchar_t * pvalue, khm_int64 * buf) {
+khc_read_int64(khm_handle pconf, const wchar_t * pvalue, khm_int64 * buf) {
     kconf_conf_space * c;
     khm_int32 rv = KHM_ERROR_SUCCESS;
 
@@ -1192,17 +1210,13 @@ khc_read_int64(khm_handle pconf, wchar_t * pvalue, khm_int64 * buf) {
         HKEY hku = NULL;
         HKEY hkm = NULL;
 
-        wchar_t * value = NULL;
+        const wchar_t * value = NULL;
         int free_space = 0;
         khm_handle conf = NULL;
 
         int i;
 
-        if(wcschr(pvalue, L'\\')
-#if 0
-           || wcschr(pvalue, L'/')
-#endif
-           ) {
+        if((value = wcsrchr(pvalue, L'\\')) != NULL) {
             if(KHM_FAILED(khc_open_space(
                 pconf, 
                 pvalue, 
@@ -1210,15 +1224,14 @@ khc_read_int64(khm_handle pconf, wchar_t * pvalue, khm_int64 * buf) {
                 &conf)))
                 goto _shadow;
             free_space = 1;
-#if 0
-            wchar_t * back, *forward;
 
-            back = wcsrchr(pvalue, L'\\');
-            forward = wcsrchr(pvalue, L'/');
-            value = (back > forward)?back:forward;
-#else
-            value = wcsrchr(pvalue, L'\\');
+            if (value) {
+                value++;
+            } else {
+#ifdef DEBUG
+                assert(FALSE);
 #endif
+            }
         } else {
             value = pvalue;
             conf = pconf;
@@ -1296,8 +1309,9 @@ _exit:
     return rv;
 }
 
+/* obtaincs cs_conf_handle/cs_conf_global */
 KHMEXP khm_int32 KHMAPI 
-khc_read_binary(khm_handle pconf, wchar_t * pvalue, 
+khc_read_binary(khm_handle pconf, const wchar_t * pvalue, 
                 void * buf, khm_size * bufsize) {
     kconf_conf_space * c;
     khm_int32 rv = KHM_ERROR_SUCCESS;
@@ -1312,15 +1326,11 @@ khc_read_binary(khm_handle pconf, wchar_t * pvalue,
         HKEY hku = NULL;
         HKEY hkm = NULL;
 
-        wchar_t * value = NULL;
+        const wchar_t * value = NULL;
         int free_space = 0;
         khm_handle conf = NULL;
 
-        if(wcschr(pvalue, L'\\')
-#if 0
-           || wcschr(pvalue, L'/')
-#endif
-           ) {
+        if((value = wcsrchr(pvalue, L'\\')) != NULL) {
             if(KHM_FAILED(khc_open_space(
                 pconf, 
                 pvalue, 
@@ -1328,15 +1338,14 @@ khc_read_binary(khm_handle pconf, wchar_t * pvalue,
                 &conf)))
                 goto _shadow;
             free_space = 1;
-#if 0
-            wchar_t * back, *forward;
 
-            back = wcsrchr(pvalue, L'\\');
-            forward = wcsrchr(pvalue, L'/');
-            value = (back > forward)?back:forward;
-#else
-            value = wcsrchr(pvalue, L'\\');
+            if (value) {
+                value++;
+            } else {
+#ifdef DEBUG
+                assert(FALSE);
 #endif
+            }
         } else {
             value = pvalue;
             conf = pconf;
@@ -1420,9 +1429,10 @@ _exit:
     return rv;
 }
 
+/* obtains cs_conf_handle/cs_conf_global */
 KHMEXP khm_int32 KHMAPI 
 khc_write_string(khm_handle pconf, 
-                 wchar_t * pvalue, 
+                 const wchar_t * pvalue, 
                  wchar_t * buf) 
 {
     HKEY pk = NULL;
@@ -1430,7 +1440,7 @@ khc_write_string(khm_handle pconf,
     khm_int32 rv = KHM_ERROR_SUCCESS;
     LONG hr;
     size_t cbsize;
-    wchar_t * value = NULL;
+    const wchar_t * value = NULL;
     int free_space = 0;
     khm_handle conf = NULL;
 
@@ -1441,27 +1451,55 @@ khc_write_string(khm_handle pconf,
     if(pconf && !khc_is_machine_handle(pconf) && !khc_is_user_handle(pconf))
         return KHM_ERROR_INVALID_OPERATION;
 
-    if(wcschr(pvalue, L'\\')
-#if 0
-       || wcschr(pvalue, L'/')
-#endif
-       ) {
-        if(KHM_FAILED(khc_open_space(
-            pconf, 
-            pvalue, 
-            KCONF_FLAG_TRAILINGVALUE | (pconf?khc_handle_flags(pconf):0), 
-            &conf)))
+    if(FAILED(StringCbLength(buf, KCONF_MAXCB_STRING, &cbsize))) {
+        rv = KHM_ERROR_INVALID_PARAM;
+        goto _exit;
+    }
+
+    cbsize += sizeof(wchar_t);
+
+    if (khc_handle_flags(pconf) & KCONF_FLAG_WRITEIFMOD) {
+        wchar_t tmpbuf[512];
+        wchar_t * buffer;
+        size_t tmpsize = cbsize;
+        khm_boolean is_equal = FALSE;
+
+        if (cbsize <= sizeof(tmpbuf)) {
+            buffer = tmpbuf;
+        } else {
+            buffer = PMALLOC(cbsize);
+        }
+
+        if (KHM_SUCCEEDED(khc_read_string(pconf, pvalue, buffer, &tmpsize)) &&
+            tmpsize == cbsize) {
+            if (khc_handle_flags(pconf) & KCONF_FLAG_IFMODCI)
+                is_equal = !_wcsicmp(buffer, buf);
+            else
+                is_equal = !wcscmp(buffer, buf);
+        }
+
+        if (buffer != tmpbuf)
+            PFREE(buffer);
+
+        if (is_equal) {
+            return KHM_ERROR_SUCCESS;
+        }
+    }
+
+    if((value = wcsrchr(pvalue, L'\\')) != NULL) {
+        if(KHM_FAILED(khc_open_space(pconf, pvalue, 
+                                     KCONF_FLAG_TRAILINGVALUE | (pconf?khc_handle_flags(pconf):0), 
+                                     &conf)))
             return KHM_ERROR_INVALID_PARAM;
         free_space = 1;
-#if 0
-        wchar_t * back, *forward;
 
-        back = wcsrchr(pvalue, L'\\');
-        forward = wcsrchr(pvalue, L'/');
-        value = (back > forward)?back:forward;
-#else
-        value = wcsrchr(pvalue, L'\\');
+        if (value) {
+            value ++;
+        } else {
+#ifdef DEBUG
+            assert(FALSE);
 #endif
+        }
     } else {
         value = pvalue;
         conf = pconf;
@@ -1475,13 +1513,6 @@ khc_write_string(khm_handle pconf,
 
     c = khc_space_from_handle(conf);
 
-    if(FAILED(StringCbLength(buf, KCONF_MAXCB_STRING, &cbsize))) {
-        rv = KHM_ERROR_INVALID_PARAM;
-        goto _exit;
-    }
-
-    cbsize += sizeof(wchar_t);
-
     if(khc_is_user_handle(conf)) {
         pk = khcint_space_open_key(c, KHM_PERM_WRITE | KHM_FLAG_CREATE);
     } else {
@@ -1499,16 +1530,17 @@ _exit:
     return rv;
 }
 
+/* obtaincs cs_conf_handle/cs_conf_global */
 KHMEXP khm_int32 KHMAPI 
 khc_write_int32(khm_handle pconf, 
-                wchar_t * pvalue, 
+                const wchar_t * pvalue, 
                 khm_int32 buf) 
 {
     HKEY pk = NULL;
     kconf_conf_space * c;
     khm_int32 rv = KHM_ERROR_SUCCESS;
     LONG hr;
-    wchar_t * value = NULL;
+    const wchar_t * value = NULL;
     int free_space = 0;
     khm_handle conf = NULL;
 
@@ -1519,11 +1551,16 @@ khc_write_int32(khm_handle pconf,
     if(pconf && !khc_is_machine_handle(pconf) && !khc_is_user_handle(pconf))
         return KHM_ERROR_INVALID_OPERATION;
 
-    if(wcschr(pvalue, L'\\')
-#if 0
-       || wcschr(pvalue, L'/')
-#endif
-       ) {
+    if (khc_handle_flags(pconf) & KCONF_FLAG_WRITEIFMOD) {
+        khm_int32 tmpvalue;
+
+        if (KHM_SUCCEEDED(khc_read_int32(pconf, pvalue, &tmpvalue)) &&
+            tmpvalue == buf) {
+            return KHM_ERROR_SUCCESS;
+        }
+    }
+
+    if((value = wcsrchr(pvalue, L'\\')) != NULL) {
         if(KHM_FAILED(khc_open_space(
             pconf, 
             pvalue, 
@@ -1531,15 +1568,14 @@ khc_write_int32(khm_handle pconf,
             &conf)))
             return KHM_ERROR_INVALID_PARAM;
         free_space = 1;
-#if 0
-        wchar_t * back, *forward;
 
-        back = wcsrchr(pvalue, L'\\');
-        forward = wcsrchr(pvalue, L'/');
-        value = (back > forward)?back:forward;
-#else
-        value = wcsrchr(pvalue, L'\\');
+        if (value) {
+            value ++;
+        } else {
+#ifdef DEBUG
+            assert(FALSE);
 #endif
+        }
     } else {
         value = pvalue;
         conf = pconf;
@@ -1568,13 +1604,14 @@ khc_write_int32(khm_handle pconf,
     return rv;
 }
 
+/* obtains cs_conf_handle/cs_conf_global */
 KHMEXP khm_int32 KHMAPI 
-khc_write_int64(khm_handle pconf, wchar_t * pvalue, khm_int64 buf) {
+khc_write_int64(khm_handle pconf, const wchar_t * pvalue, khm_int64 buf) {
     HKEY pk = NULL;
     kconf_conf_space * c;
     khm_int32 rv = KHM_ERROR_SUCCESS;
     LONG hr;
-    wchar_t * value = NULL;
+    const wchar_t * value = NULL;
     int free_space = 0;
     khm_handle conf = NULL;
 
@@ -1585,11 +1622,16 @@ khc_write_int64(khm_handle pconf, wchar_t * pvalue, khm_int64 buf) {
     if(pconf && !khc_is_machine_handle(pconf) && !khc_is_user_handle(pconf))
         return KHM_ERROR_INVALID_OPERATION;
 
-    if(wcschr(pvalue, L'\\')
-#if 0
-       || wcschr(pvalue, L'/')
-#endif
-       ) {
+    if (khc_handle_flags(pconf) & KCONF_FLAG_WRITEIFMOD) {
+        khm_int64 tmpvalue;
+
+        if (KHM_SUCCEEDED(khc_read_int64(pconf, pvalue, &tmpvalue)) &&
+            tmpvalue == buf) {
+            return KHM_ERROR_SUCCESS;
+        }
+    }
+
+    if((value = wcsrchr(pvalue, L'\\')) != NULL) {
         if(KHM_FAILED(khc_open_space(
             pconf, 
             pvalue, 
@@ -1597,15 +1639,14 @@ khc_write_int64(khm_handle pconf, wchar_t * pvalue, khm_int64 buf) {
             &conf)))
             return KHM_ERROR_INVALID_PARAM;
         free_space = 1;
-#if 0
-        wchar_t * back, *forward;
 
-        back = wcsrchr(pvalue, L'\\');
-        forward = wcsrchr(pvalue, L'/');
-        value = (back > forward)?back:forward;
-#else
-        value = wcsrchr(pvalue, L'\\');
+        if (value) {
+            value ++;
+        } else {
+#ifdef DEBUG
+            assert(FALSE);
 #endif
+        }
     } else {
         value = pvalue;
         conf = pconf;
@@ -1634,15 +1675,16 @@ khc_write_int64(khm_handle pconf, wchar_t * pvalue, khm_int64 buf) {
     return rv;
 }
 
+/* obtains cs_conf_handle/cs_conf_global */
 KHMEXP khm_int32 KHMAPI 
 khc_write_binary(khm_handle pconf, 
-                 wchar_t * pvalue, 
+                 const wchar_t * pvalue, 
                  void * buf, khm_size bufsize) {
     HKEY pk = NULL;
     kconf_conf_space * c;
     khm_int32 rv = KHM_ERROR_SUCCESS;
     LONG hr;
-    wchar_t * value = NULL;
+    const wchar_t * value = NULL;
     int free_space = 0;
     khm_handle conf = NULL;
 
@@ -1653,11 +1695,7 @@ khc_write_binary(khm_handle pconf,
     if(pconf && !khc_is_machine_handle(pconf) && !khc_is_user_handle(pconf))
         return KHM_ERROR_INVALID_OPERATION;
 
-    if(wcschr(pvalue, L'\\')
-#if 0
-       || wcschr(pvalue, L'/')
-#endif
-       ) {
+    if((value = wcsrchr(pvalue, L'\\')) != NULL) {
         if(KHM_FAILED(khc_open_space(
             pconf, 
             pvalue, 
@@ -1665,15 +1703,14 @@ khc_write_binary(khm_handle pconf,
             &conf)))
             return KHM_ERROR_INVALID_PARAM;
         free_space = 1;
-#if 0
-        wchar_t * back, *forward;
 
-        back = wcsrchr(pvalue, L'\\');
-        forward = wcsrchr(pvalue, L'/');
-        value = (back > forward)?back:forward;
-#else
-        value = wcsrchr(pvalue, L'\\');
+        if (value) {
+            value ++;
+        } else {
+#ifdef DEBUG
+            assert(FALSE);
 #endif
+        }
     } else {
         value = pvalue;
         conf = pconf;
@@ -1702,6 +1739,7 @@ khc_write_binary(khm_handle pconf,
     return rv;
 }
 
+/* no locks */
 KHMEXP khm_int32 KHMAPI 
 khc_get_config_space_name(khm_handle conf, 
                           wchar_t * buf, khm_size * bufsize) {
@@ -1743,6 +1781,7 @@ khc_get_config_space_name(khm_handle conf,
     return rv;
 }
 
+/* obtains cs_conf_handle/cs_conf_global */
 KHMEXP khm_int32 KHMAPI 
 khc_get_config_space_parent(khm_handle conf, khm_handle * parent) {
     kconf_conf_space * c;
@@ -1763,8 +1802,9 @@ khc_get_config_space_parent(khm_handle conf, khm_handle * parent) {
     return KHM_ERROR_SUCCESS;
 }
 
+/* obtains cs_conf_global */
 KHMEXP khm_int32 KHMAPI 
-khc_get_type(khm_handle conf, wchar_t * value) {
+khc_get_type(khm_handle conf, const wchar_t * value) {
     HKEY hkm = NULL;
     HKEY hku = NULL;
     kconf_conf_space * c;
@@ -1778,7 +1818,7 @@ khc_get_type(khm_handle conf, wchar_t * value) {
     if(!khc_is_handle(conf))
         return KC_NONE;
 
-    c = (kconf_conf_space *) conf;
+    c = khc_space_from_handle(conf);
 
     if(!khc_is_machine_handle(conf))
         hku = khcint_space_open_key(c, KHM_PERM_READ);
@@ -1821,8 +1861,9 @@ khc_get_type(khm_handle conf, wchar_t * value) {
     return rv;
 }
 
+/* obtains cs_conf_global */
 KHMEXP khm_int32 KHMAPI 
-khc_value_exists(khm_handle conf, wchar_t * value) {
+khc_value_exists(khm_handle conf, const wchar_t * value) {
     HKEY hku = NULL;
     HKEY hkm = NULL;
     kconf_conf_space * c;
@@ -1860,8 +1901,9 @@ khc_value_exists(khm_handle conf, wchar_t * value) {
     return rv;
 }
 
+/* obtains cs_conf_global */
 KHMEXP khm_int32 KHMAPI
-khc_remove_value(khm_handle conf, wchar_t * value, khm_int32 flags) {
+khc_remove_value(khm_handle conf, const wchar_t * value, khm_int32 flags) {
     HKEY hku = NULL;
     HKEY hkm = NULL;
     kconf_conf_space * c;
@@ -1981,6 +2023,7 @@ khcint_remove_space(kconf_conf_space * c, khm_int32 flags) {
     return KHM_ERROR_SUCCESS;
 }
 
+/* obtains cs_conf_global */
 KHMEXP khm_int32 KHMAPI
 khc_remove_space(khm_handle conf) {
 
@@ -2023,6 +2066,7 @@ khc_remove_space(khm_handle conf) {
     return rv;
 }
 
+/* no locks */
 khm_boolean 
 khcint_is_valid_name(wchar_t * name)
 {
@@ -2032,8 +2076,9 @@ khcint_is_valid_name(wchar_t * name)
     return TRUE;
 }
 
+/* no locks */
 khm_int32 
-khcint_validate_schema(kconf_schema * schema,
+khcint_validate_schema(const kconf_schema * schema,
                        int begin,
                        int *end)
 {
@@ -2094,8 +2139,9 @@ khcint_validate_schema(kconf_schema * schema,
     return KHM_ERROR_SUCCESS;
 }
 
+/* obtains cs_conf_handle/cs_conf_global; called with cs_conf_global */
 khm_int32 
-khcint_load_schema_i(khm_handle parent, kconf_schema * schema, 
+khcint_load_schema_i(khm_handle parent, const kconf_schema * schema, 
                      int begin, int * end)
 {
     int i;
@@ -2108,11 +2154,16 @@ khcint_load_schema_i(khm_handle parent, kconf_schema * schema,
     while(!end_found) {
         switch(state) {
             case 0: /* initial.  this record should start a config space */
+                LeaveCriticalSection(&cs_conf_global);
                 if(KHM_FAILED(khc_open_space(parent, schema[i].name, 
-                                             KHM_FLAG_CREATE, &h)))
+                                             KHM_FLAG_CREATE, &h))) {
+                    EnterCriticalSection(&cs_conf_global);
                     return KHM_ERROR_INVALID_PARAM;
+                }
+                EnterCriticalSection(&cs_conf_global);
                 thisconf = khc_space_from_handle(h);
                 thisconf->schema = schema + (begin + 1);
+                thisconf->nSchema = 0;
                 state = 1;
                 break;
 
@@ -2127,7 +2178,9 @@ khcint_load_schema_i(khm_handle parent, kconf_schema * schema,
                     end_found = 1;
                     if(end)
                         *end = i;
+                    LeaveCriticalSection(&cs_conf_global);
                     khc_close_space(h);
+                    EnterCriticalSection(&cs_conf_global);
                 }
                 break;
 
@@ -2139,7 +2192,9 @@ khcint_load_schema_i(khm_handle parent, kconf_schema * schema,
                     end_found = 1;
                     if(end)
                         *end = i;
+                    LeaveCriticalSection(&cs_conf_global);
                     khc_close_space(h);
+                    EnterCriticalSection(&cs_conf_global);
                 } else {
                     return KHM_ERROR_INVALID_PARAM;
                 }
@@ -2155,8 +2210,9 @@ khcint_load_schema_i(khm_handle parent, kconf_schema * schema,
     return KHM_ERROR_SUCCESS;
 }
 
+/* obtains cs_conf_handle/cs_conf_global */
 KHMEXP khm_int32 KHMAPI 
-khc_load_schema(khm_handle conf, kconf_schema * schema)
+khc_load_schema(khm_handle conf, const kconf_schema * schema)
 {
     khm_int32 rv = KHM_ERROR_SUCCESS;
 
@@ -2176,8 +2232,9 @@ khc_load_schema(khm_handle conf, kconf_schema * schema)
     return rv;
 }
 
+/* obtains cs_conf_handle/cs_conf_global; called with cs_conf_global */
 khm_int32 
-khcint_unload_schema_i(khm_handle parent, kconf_schema * schema, 
+khcint_unload_schema_i(khm_handle parent, const kconf_schema * schema, 
                        int begin, int * end)
 {
     int i;
@@ -2190,8 +2247,12 @@ khcint_unload_schema_i(khm_handle parent, kconf_schema * schema,
     while(!end_found) {
         switch(state) {
             case 0: /* initial.  this record should start a config space */
-                if(KHM_FAILED(khc_open_space(parent, schema[i].name, 0, &h)))
+                LeaveCriticalSection(&cs_conf_global);
+                if(KHM_FAILED(khc_open_space(parent, schema[i].name, 0, &h))) {
+                    EnterCriticalSection(&cs_conf_global);
                     return KHM_ERROR_INVALID_PARAM;
+                }
+                EnterCriticalSection(&cs_conf_global);
                 thisconf = khc_space_from_handle(h);
                 if(thisconf->schema == (schema + (begin + 1))) {
                     thisconf->schema = NULL;
@@ -2209,7 +2270,9 @@ khcint_unload_schema_i(khm_handle parent, kconf_schema * schema,
                     end_found = 1;
                     if(end)
                         *end = i;
+                    LeaveCriticalSection(&cs_conf_global);
                     khc_close_space(h);
+                    EnterCriticalSection(&cs_conf_global);
                 }
                 break;
 
@@ -2221,7 +2284,9 @@ khcint_unload_schema_i(khm_handle parent, kconf_schema * schema,
                     end_found = 1;
                     if(end)
                         *end = i;
+                    LeaveCriticalSection(&cs_conf_global);
                     khc_close_space(h);
+                    EnterCriticalSection(&cs_conf_global);
                 } else {
                     return KHM_ERROR_INVALID_PARAM;
                 }
@@ -2237,8 +2302,9 @@ khcint_unload_schema_i(khm_handle parent, kconf_schema * schema,
     return KHM_ERROR_SUCCESS;
 }
 
+/* obtains cs_conf_handle/cs_conf_global */
 KHMEXP khm_int32 KHMAPI 
-khc_unload_schema(khm_handle conf, kconf_schema * schema)
+khc_unload_schema(khm_handle conf, const kconf_schema * schema)
 {
     khm_int32 rv = KHM_ERROR_SUCCESS;
 
@@ -2258,6 +2324,7 @@ khc_unload_schema(khm_handle conf, kconf_schema * schema)
     return rv;
 }
 
+/* obtaincs cs_conf_handle/cs_conf_global */
 KHMEXP khm_int32 KHMAPI 
 khc_enum_subspaces(khm_handle conf,
                    khm_handle prev,
@@ -2365,8 +2432,9 @@ khc_enum_subspaces(khm_handle conf,
     return rv;
 }
 
+/* obtains cs_conf_handle/cs_conf_global */
 KHMEXP khm_int32 KHMAPI 
-khc_write_multi_string(khm_handle conf, wchar_t * value, wchar_t * buf)
+khc_write_multi_string(khm_handle conf, const wchar_t * value, wchar_t * buf)
 {
     size_t cb;
     wchar_t vbuf[KCONF_MAXCCH_STRING];
@@ -2396,8 +2464,9 @@ khc_write_multi_string(khm_handle conf, wchar_t * value, wchar_t * buf)
     return rv;
 }
 
+/* obtains cs_conf_handle/cs_conf_global */
 KHMEXP khm_int32 KHMAPI 
-khc_read_multi_string(khm_handle conf, wchar_t * value, 
+khc_read_multi_string(khm_handle conf, const wchar_t * value, 
                       wchar_t * buf, khm_size * bufsize)
 {
     wchar_t vbuf[KCONF_MAXCCH_STRING];
