@@ -106,9 +106,9 @@ khm_convert524(krb5_context alt_ctx)
     increds.times.endtime = 0;
     increds.keyblock.enctype = ENCTYPE_DES_CBC_CRC;
     if ((code = pkrb5_get_credentials(ctx, 0,
-        cc,
-        &increds,
-        &v5creds))) 
+                                      cc,
+                                      &increds,
+                                      &v5creds))) 
     {
         goto cleanup;
     }
@@ -1123,6 +1123,8 @@ khm_krb5_kinit(krb5_context       alt_ctx,
     _reportf(L"In khm_krb5_kinit");
 
     pkrb5_get_init_creds_opt_init(&options);
+    pkrb5_get_init_creds_opt_set_change_password_prompt(&options, 0);
+
     memset(&my_creds, 0, sizeof(my_creds));
 
     if (alt_ctx) {
@@ -2108,6 +2110,27 @@ khm_krb5_get_temp_profile_file(LPSTR confname, UINT szConfname)
     return FALSE;
 }
 
+#ifdef NOT_QUITE_IMPLEMENTED_YET
+BOOL
+khm_krb5_set_profile_file(krb5_context ctx, LPSTR confname)
+{
+    char *conffiles[2];
+
+    if (confname == NULL ||
+        pkrb5_set_config_files == NULL ||
+        ctx == NULL)
+        return FALSE;
+
+    conffiles[0] = confname;
+    conffiles[1] = NULL;
+
+    if (pkrb5_set_config_files(ctx, conffiles))
+        return FALSE;
+    else
+        return TRUE;
+}
+#endif
+
 BOOL 
 khm_krb5_get_profile_file(LPSTR confname, UINT szConfname)
 {
